package com.mobage.nbpf.sampleapp.recovery;

import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.util.Base64;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.RequestEntity;
import org.springframework.http.ResponseEntity;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.HttpServerErrorException;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

import com.mobage.nbpf.sampleapp.auth.MobageTokenEndpointResponse;
import com.mobage.nbpf.sampleapp.config.SampleAppConfig;
import com.mobage.nbpf.sampleapp.exception.MobageException;
import com.mobage.nbpf.sampleapp.exception.MobageUnauthorizedException;

/**
 * Service class that provides communication to Mobage Bank Debit APIs.
 */
@Service
public class PaymentRecoveryService {
  Logger logger = LoggerFactory.getLogger(PaymentRecoveryService.class);

  private SampleAppConfig config;
  private JdbcTemplate jdbcTemplate;

  @Autowired
  public PaymentRecoveryService(SampleAppConfig config, JdbcTemplate jdbcTemplate) {
    this.config = config;
    this.jdbcTemplate = jdbcTemplate;
  }

  /**
   * Send a request to Mobage Server and get the payment process.
   */
  private PaymentBankDebitResponse getBankDebit(String transactionId, String accessToken) {

    // Set authorization header with Access Token
    HttpHeaders headers = new HttpHeaders();
    headers.add("Authorization", "Bearer " + accessToken);
    headers.add("Accept", "*/*");

    // Send a request
    PaymentBankDebitResponse response;
    RestTemplate restTemplate = new RestTemplate();
    try {
      String url = config.getBankDebitEndpoint() + "/" + transactionId;
      ResponseEntity<PaymentBankDebitResponse> responseEntity =
          restTemplate.exchange(
              RequestEntity.get(URI.create(url)).headers(headers).build(),
              PaymentBankDebitResponse.class);
      response = responseEntity.getBody();
    } catch (HttpClientErrorException e) {

      // In this case, the error is caused by Invalid Request
      // Status Code 400； The error is caused by invalid parameter in JSON Request Body
      // Status Code 401； The error is caused by invalid HTTP Authorization Header
      //                   Please make sure if Authorization Header is valid.
      logger.error(e.getResponseBodyAsString(),e);
      if (e.getStatusCode() == HttpStatus.UNAUTHORIZED) {
        throw new MobageUnauthorizedException(
            "Failed to communitate to Mobage because valid access token is not set");
      } else {
        throw new MobageException(
            "Failed to communicate to Mobage because of an invalid request. Status Code:"
                + e.getStatusCode());
      }
    } catch (HttpServerErrorException e) {
      // In this case, the error might be caused by Mobage Platform
      throw new MobageException("Unexpected error occurred on Mobage Platform");
    } catch (RestClientException e) {
      // Typically in case of error while extracting response
      throw new MobageException(e.getMessage());
    }
    return response;
  }

  /** Give items and close order process. */
  @Transactional
  private void giveOrderItem(String orderId, String itemId, int quantity) {

    Map<String, Object> order =
        jdbcTemplate.queryForMap("select * from order_db where order_id = ? for update", orderId);

    if (!order.get("order_state").toString().equals("authorized")) {
      throw new MobageException("order_state is not authorized.");
    }
    String userId = order.get("user_id").toString();
    jdbcTemplate.update(
        "insert into user_items"
            + " (user_id, item_id, item_num) values (?, ?, ?)"
            + " on duplicate key update item_num = item_num + ?;",
        userId,
        itemId,
        quantity,
        quantity);

    jdbcTemplate.update("update order_db set order_state = ? where order_id = ?", "close", orderId);
  }

  /** cancel order process. */
  @Transactional
  private void cancelOrderItem(String orderId, String state) {
    jdbcTemplate.update("update order_db set order_state = ? where order_id = ?;", state, orderId);
  }

  /**
   * check authorized orders some time passed and do remaining processing.
   */
  public void recovery() {
    List<Map<String, Object>> list =
        jdbcTemplate.queryForList(
            "select a.order_id, a.user_id, a.client_id, a.transaction_id, a.order_state, b.refresh_token"
                + " from order_db a left join user_tokens b on (a.user_id = b.user_id and a.client_id = b.client_id)"
                + " where a.client_id = ? and  a.order_state = 'authorized' and a.created_at < ? order by a.created_at;",
            config.getClientId(),
            System.currentTimeMillis() / 1000 - (30 * 60));
    for (Map<String, Object> record : list) {
      String orderId = record.get("order_id").toString();
      String userId = record.get("user_id").toString();
      String transactionId = record.get("transaction_id").toString();
      String orderState = record.get("transaction_id").toString();
      String refreshToken = record.get("refresh_token").toString();

      MobageTokenEndpointResponse authToken = requestAccessToken(refreshToken);
      String accessToken = authToken.getAccessToken();

      // if the new refresh_token is different from previous one,
      // you have to update the refresh_token on your DataBase.
      if (!refreshToken.equals(authToken.getRefreshToken())) {
        Long refreshTokenExpiresAt = System.currentTimeMillis() / 1000 + (90 * 24 * 3600);
        saveRefreshToken(userId, authToken.getRefreshToken(), refreshTokenExpiresAt);
      }


      PaymentBankDebitResponse res = getBankDebit(transactionId, accessToken);
      String state = res.getState();
      if (state == null) {
        logger.error(
            String.format(
                "Failure. Not Found. {order_id: %s, user_id: %s, %s}",
                orderId, userId, res.getItems().get(0).getItem().getId()));
      } else if (orderState.equals(state)) {
        logger.error(
            String.format(
                "Failure. order_db order_state is already syncronized. state=%s {order_id: %s, user_id: %s}",
                orderState, orderId, userId));
      } else if (state.equals("canceled") || state.equals("error")) {
        cancelOrderItem(orderId, state);
        logger.info(String.format("Canceled. {order_id: %s, user_id: %s}", orderId, userId));
      } else if (state.equals("closed")) {
        String itemId = res.getItems().get(0).getItem().getId();
        int quantity = res.getItems().get(0).getQuantity();
        giveOrderItem(
            orderId, res.getItems().get(0).getItem().getId(), res.getItems().get(0).getQuantity());
        logger.info(
            String.format(
                "Success. {order_id: %s, user_id: %s, item_id: %s, quantity: %d}",
                orderId, userId, itemId, quantity));
      }
    }
  }

  /**
   * request an Access Token for Mobage.
   */
  private MobageTokenEndpointResponse requestAccessToken(String refreshToken) {
    HttpHeaders headers = new HttpHeaders();
    headers.setContentType(MediaType.APPLICATION_FORM_URLENCODED);

    // Set an authorization header with credentials
    String credential = config.getClientId() + ":" + config.getClientSecret();
    String encoded =
        Base64.getEncoder().encodeToString(credential.getBytes(StandardCharsets.UTF_8));
    headers.add("Authorization", "Basic " + encoded);

    MultiValueMap<String, String> map = new LinkedMultiValueMap<>();
    map.add("grant_type", "refresh_token");
    map.add("redirect_uri", config.getRedirectUri());
    map.add("client_id", config.getClientId());
    map.add("client_secret", config.getClientSecret());
    map.add("refresh_token", refreshToken);

    // Create a POST Request
    RequestEntity<MultiValueMap<String, String>> requestEntity =
        RequestEntity.post(URI.create(config.getTokenEndpoint()))
            .contentType(MediaType.APPLICATION_FORM_URLENCODED)
            .headers(headers)
            .body(map);

    // Send a request
    MobageTokenEndpointResponse token = null;
    RestTemplate restTemplate = new RestTemplate();
    try {
      ResponseEntity<MobageTokenEndpointResponse> response =
          restTemplate.exchange(requestEntity, MobageTokenEndpointResponse.class);
      token = response.getBody();
    } catch (HttpClientErrorException e) {
      // In this case, the error is caused by Invalid Request
      // Status Code 400； The error is caused by invalid parameter in JSON Request Body
      //                   Please make sure if redirect_uri parameter is exactly same as
      //                   the one passed to JavaScript SDK
      // Status Code 401； The error is caused by invalid HTTP Authorization Header
      //                   Please make sure if Authorization Header is valid.
      //                   Authorization header needs to be generated from the secrets for the app
      //                   registered on Mobage Developer Portal
      throw new MobageException("Failed to get access token. Status Code:" + e.getStatusCode());
    } catch (HttpServerErrorException e) {
      // In this case, the error might be caused by Mobage Platform
      throw new MobageException("Unexpected error occurred on Mobage Platform");
    }
    return token;
  }

  /**
   * store the Request Token in Database.
   */
  private void saveRefreshToken(String userId, String refreshToken, Long expiresAt) {
    jdbcTemplate.update(
        "insert into user_tokens"
            + " (user_id, client_id, refresh_token, expires_at) values (?, ?, ?, ?)"
            + " on duplicate key update refresh_token = ?, expires_at = ?;",
        userId,
        config.getClientId(),
        refreshToken,
        expiresAt,
        refreshToken,
        expiresAt);
  }
}
