package com.mobage.nbpf.sampleapp.auth;

import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.X509EncodedKeySpec;
import java.util.Base64;
import java.util.Date;
import java.util.Map;
import java.io.IOException;
import javax.servlet.http.HttpServletResponse;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.ResponseStatus;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.bind.annotation.RequestParam;

import com.mobage.nbpf.sampleapp.SessionStorage;
import com.mobage.nbpf.sampleapp.config.SampleAppConfig;
import com.mobage.nbpf.sampleapp.exception.MobageException;

import io.jsonwebtoken.Claims;
import io.jsonwebtoken.JwtException;
import io.jsonwebtoken.Jwts;

@RestController
@RequestMapping("/mobage/auth/callback")
public class MobageAuthCallbackController {
  private MobageAuthService authService;
  private SampleAppConfig config;
  private SessionStorage session;
  Logger logger = LoggerFactory.getLogger(MobageAuthCallbackController.class);

  /**
   * Constructor.
   */
  @Autowired
  public MobageAuthCallbackController(MobageAuthService authService,
                                      SampleAppConfig config,
                                      SessionStorage session) {
    this.authService = authService;
    this.config = config;
    this.session = session;
  }

  /**
   * request handler for authorization callback URI.
   */
  @PostMapping
  @ResponseStatus(HttpStatus.OK)
  public MobageAuthCallbackResponse callbackPost(
      @RequestBody MobageAuthCallbackRequest request, HttpServletResponse response) {
    doCallbackCommon(request.getCode(), request.getState(), request.getSessionState());
    return new MobageAuthCallbackResponse(true, session.getUserId());
  }

  @GetMapping
  public void callbackGet(@RequestParam Map<String, String> params, HttpServletResponse response)
      throws IOException {
    doCallbackCommon(params.get("code"), params.get("state"), params.get("session_state"));
    response.sendRedirect("/mobage/auth");
  }

  private void doCallbackCommon(String code, String state, String sessionState) {
    MobageTokenEndpointResponse authToken = authService.requestAccessToken(code);

    // Validate if state parameter from request body is equal to state parameter from session cookie
    if (!session.getState().equals(state)) {
      throw new MobageException("state parameter is invalid");
    }

    String userId;
    Claims claims;
    try {
      // Validate Access Token with public key of Mobage Platform
      byte[] decodedKey = Base64.getDecoder().decode(config.getPubkey());
      KeyFactory keyFactory = KeyFactory.getInstance("RSA");
      PublicKey pubKey = keyFactory.generatePublic(new X509EncodedKeySpec(decodedKey));
      String idToken = authToken.getIdToken();
      claims = Jwts.parser().setSigningKey(pubKey).parseClaimsJws(idToken).getBody();

      validateClaims(claims);
      // Mobage User ID can be obtained from subject
      userId = claims.getSubject();

    } catch (JwtException ex) {
      throw new MobageException("JWT is invalid: underlying error: " + ex.getMessage());
    } catch (NoSuchAlgorithmException | InvalidKeySpecException ex) {
      throw new MobageException("Internal Error: underlying error: " + ex.getMessage());
    }

    // Set a session_state parameter into session cookie
    // This session_state should be passed to mobage.event.subscribe('oauth.sessionStateChange')
    // to receive notification for logout event on Mobage
    session.setSessionState(sessionState);

    session.setAccessToken(authToken.getAccessToken());
    session.setRefreshToken(authToken.getRefreshToken());
    session.setUserId(userId);

    Long refreshTokenExpiresAt = System.currentTimeMillis() / 1000 + (90 * 24 * 3600);
    authService.saveRefreshToken(userId, authToken.getRefreshToken(), refreshTokenExpiresAt);

    logger.info(String.format("Authorization process is successful. {user_id: %s}", userId));
  }

  private void validateClaims(Claims claims) {
    if (!claims.getIssuer().equals(config.getIssuer())) {
      throw new MobageException("Issuer Claim is invalid.");
    }

    if (!claims.getAudience().equals(config.getClientId())) {
      throw new MobageException("Audience Claim is invalid.");
    }

    Date now = new Date();
    if (claims.getIssuedAt().after(now)) {
      throw new MobageException("Issued At Claim is invalid.");
    }

    if (claims.getExpiration().before(now)) {
      throw new MobageException("This JWT is expired.");
    }
  }


}
