package com.mobage.nbpf.sampleapp.auth;

import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.util.Base64;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.RequestEntity;
import org.springframework.http.ResponseEntity;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.stereotype.Service;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.HttpServerErrorException;
import org.springframework.web.client.RestTemplate;

import com.mobage.nbpf.sampleapp.SessionStorage;
import com.mobage.nbpf.sampleapp.config.SampleAppConfig;
import com.mobage.nbpf.sampleapp.exception.MobageException;

/**
 * Service class that provides functionality of communicating with Mobage Platform
 * for Authorization.
 */
@Service
public class MobageAuthService {

  private SampleAppConfig config;
  private RestTemplate restTemplate;
  private JdbcTemplate jdbcTemplate;

  @Autowired
  public MobageAuthService(
      RestTemplate restTemplate,
      SampleAppConfig config,
      SessionStorage session,
      JdbcTemplate jdbcTemplate) {
    this.restTemplate = restTemplate;
    this.config = config;
    this.jdbcTemplate = jdbcTemplate;
  }

  /**
   * request an Access Token for Mobage.
   */
  public MobageTokenEndpointResponse requestAccessToken(String code) {
    HttpHeaders headers = new HttpHeaders();
    headers.setContentType(MediaType.APPLICATION_FORM_URLENCODED);

    // Set an authorization header with credentials
    String credential = config.getClientId() + ":" + config.getClientSecret();
    String encoded =
        Base64.getEncoder().encodeToString(credential.getBytes(StandardCharsets.UTF_8));
    headers.add("Authorization", "Basic " + encoded);

    MultiValueMap<String, String> map = new LinkedMultiValueMap<>();
    map.add("client_id", config.getClientId());
    map.add("redirect_uri", config.getRedirectUri());
    map.add("code", code);
    map.add("grant_type", "authorization_code");

    // Create a POST Request
    RequestEntity<MultiValueMap<String, String>> requestEntity =
        RequestEntity.post(URI.create(config.getTokenEndpoint()))
            .contentType(MediaType.APPLICATION_FORM_URLENCODED)
            .headers(headers)
            .body(map);

    // Send a request
    MobageTokenEndpointResponse token;
    try {
      ResponseEntity<MobageTokenEndpointResponse> response =
          this.restTemplate.exchange(requestEntity, MobageTokenEndpointResponse.class);
      token = response.getBody();
    } catch (HttpClientErrorException e) {
      // In this case, the error is caused by Invalid Request
      // Status Code 400； The error is caused by invalid parameter in JSON Request Body
      //                   Please make sure if redirect_uri parameter is exactly same as
      //                   the one passed to JavaScript SDK
      // Status Code 401； The error is caused by invalid HTTP Authorization Header
      //                   Please make sure if Authorization Header is valid.
      //                   Authorization header needs to be generated from the secrets for the app
      //                   registered on Mobage Developer Portal
      throw new MobageException("Failed to get access token. Status Code:" + e.getStatusCode());
    } catch (HttpServerErrorException e) {
      // In this case, the error might be caused by Mobage Platform
      throw new MobageException("Unexpected error occurred on Mobage Platform");
    }
    return token;
  }

  public void saveRefreshToken(String userId, String refreshToken, Long expiresAt) {
    jdbcTemplate.update(
        "insert into user_tokens"
            + " (user_id, client_id, refresh_token, expires_at) values (?, ?, ?, ?)"
            + " on duplicate key update refresh_token = ?, expires_at = ?;",
        userId,
        config.getClientId(),
        refreshToken,
        expiresAt,
        refreshToken,
        expiresAt);
  }
}
