package com.mobage.nbpf.sampleapp.payment;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.mobage.nbpf.sampleapp.SessionStorage;
import com.mobage.nbpf.sampleapp.config.SampleAppConfig;
import com.mobage.nbpf.sampleapp.exception.MobageException;

import io.jsonwebtoken.Claims;
import io.jsonwebtoken.JwtException;
import io.jsonwebtoken.Jwts;
import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.X509EncodedKeySpec;
import java.util.Base64;
import java.util.Date;
import javax.servlet.http.HttpServletResponse;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.web.bind.annotation.*;


@RestController
@RequestMapping("/api/payment")
public class MobagePaymentRestController {
  private MobagePaymentService service;
  private SampleAppConfig config;
  private SessionStorage session;
  Logger logger = LoggerFactory.getLogger(MobagePaymentRestController.class);

  /**
   * Constructor.
   */
  @Autowired
  public MobagePaymentRestController(MobagePaymentService service,
                                     SampleAppConfig config,
                                     SessionStorage session) {
    this.service = service;
    this.config = config;
    this.session = session;
  }

  /**
   * request handler for start payment transaction.
   */
  @PostMapping("/start")
  @ResponseStatus(HttpStatus.OK)
  public MobagePaymentStartResponse startPayment(
      @RequestBody MobagePaymentStartRequest request, HttpServletResponse response) {

    MobageBankDebitPostResponse res =
        service.postBankDebit(session.getAccessToken(), request.getItemId(), request.getQuantity());

    // order Id should be autogenerated unique ID that represents identifier of payment transaction
    // in your system
    String orderId = service.publishOrderId(session.getUserId(), res.getId());

    return new MobagePaymentStartResponse(res.getId(), orderId);
  }


  /**
   * request handler for confirming receipt.
   */
  @PostMapping("/confirm")
  @ResponseStatus(HttpStatus.OK)
  public String confirmReceipt(@RequestBody String requestBody) {

    Claims claims = validateReceipt(requestBody);
    validateClaims(claims);

    ObjectMapper mapper = new ObjectMapper();
    MobagePaymentCustomClaim.Extra claimExtra =
        mapper.convertValue(claims.get("extra"), MobagePaymentCustomClaim.Extra.class);

    // Give virtual item to the user
    service.giveOrderItem(
        claimExtra.getResult().getOrderId(),
        claimExtra.getResult().getPayment().getItems().get(0).getItem().getId(),
        claimExtra.getResult().getPayment().getItems().get(0).getQuantity());
    return "OK";
  }

  /**
   * request handler for cancel.
   */
  @PostMapping("/cancel")
  @ResponseStatus(HttpStatus.OK)
  public String cancel(@RequestBody MobagePaymentCancelRequest request) {

    service.cancelOrder(request.getOrderId(), request.getTransactionId(), session.getAccessToken());
    return "OK";
  }

  private Claims validateReceipt(String receipt) {
    Claims claims;
    try {
      // Validate Access Token with public key of Mobage Platform
      byte[] decodedKey = Base64.getDecoder().decode(config.getPubkey());
      KeyFactory keyFactory = KeyFactory.getInstance("RSA");
      PublicKey pubKey = keyFactory.generatePublic(new X509EncodedKeySpec(decodedKey));
      claims = Jwts.parser().setSigningKey(pubKey).parseClaimsJws(receipt).getBody();

      return claims;

    } catch (JwtException ex) {
      throw new MobageException("JWT is invalid: underlying error: " + ex.getMessage());
    } catch (NoSuchAlgorithmException | InvalidKeySpecException ex) {
      throw new MobageException("Internal Error: underlying error: " + ex.getMessage());
    }
  }

  private void validateClaims(Claims claims) {
    // Validate Issuer Claim
    if (!claims.getIssuer().equals(config.getIssuerForWidget())) {
      throw new MobageException("Issuer Claim is invalid.");
    }

    // Validate Audience Claim
    if (!claims.getAudience().equals(config.getClientId())) {
      throw new MobageException("Audience Claim is invalid.");
    }

    // Validate IssuedAt Claim
    Date now = new Date();
    if (claims.getIssuedAt().after(now)) {
      throw new MobageException("Issued At Claim is invalid.");
    }

    // Validate User ID
    if (!claims.getSubject().equals(session.getUserId())) {
      throw new MobageException("Subject Claim is invalid.");
    }

    // Validate if payment state is closed
    ObjectMapper mapper = new ObjectMapper();
    MobagePaymentCustomClaim.Extra extra =
        mapper.convertValue(claims.get("extra"), MobagePaymentCustomClaim.Extra.class);
    if (!extra.getResult().getPayment().getState().equals("closed")) {
      throw new MobageException("Payment state is not closed.");
    }
  }
}
