package com.mobage.nbpf.sampleapp.payment;

import com.mobage.nbpf.sampleapp.SessionStorage;
import com.mobage.nbpf.sampleapp.config.SampleAppConfig;
import com.mobage.nbpf.sampleapp.exception.MobageException;
import com.mobage.nbpf.sampleapp.exception.MobageUnauthorizedException;

import java.net.URI;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.*;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.PreparedStatementCreator;
import org.springframework.jdbc.support.GeneratedKeyHolder;
import org.springframework.jdbc.support.KeyHolder;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.HttpServerErrorException;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

/**
 * Service class that provides communication to Mobage Bank Debit APIs.
 */
@Service
public class MobagePaymentService {
  Logger logger = LoggerFactory.getLogger(MobagePaymentService.class);

  private SampleAppConfig config;
  private RestTemplate restTemplate;
  private JdbcTemplate jdbcTemplate;

  @Autowired
  public MobagePaymentService(
      RestTemplate restTemplate,
      SampleAppConfig config,
      SessionStorage session,
      JdbcTemplate jdbcTemplate) {
    this.restTemplate = restTemplate;
    this.config = config;
    this.jdbcTemplate = jdbcTemplate;
  }

  /**
   * Send a request to Mobage Server and initiate a payment process.
   */
  public MobageBankDebitPostResponse postBankDebit(
      String accessToken, String itemId, int quantity) {

    // Create a POJO for Request Body
    MobageBankDebitPostRequest.ItemProperty prop = MobageBankDebitPostRequest.ItemProperty.builder()
            .id(itemId)
            .build();
    MobageBankDebitPostRequest.Item item =
        MobageBankDebitPostRequest.Item.builder()
            .item(prop)
            .quantity(quantity)
            .build();
    List<MobageBankDebitPostRequest.Item> items = Arrays.asList(item);
    MobageBankDebitPostRequest request = MobageBankDebitPostRequest.builder()
            .items(items)
            .comment("Item description. The comment here will be displayed "
                    + "on the user's payment history screen")
            .state("authorized")
            .build();

    // Set authorization header with Access Token
    HttpHeaders headers = new HttpHeaders();
    headers.add("Authorization", "Bearer " + accessToken);

    // Create a POST Request
    RequestEntity<MobageBankDebitPostRequest> requestEntity =
        RequestEntity.post(URI.create(config.getBankDebitEndpoint()))
            .contentType(MediaType.APPLICATION_JSON)
            .headers(headers)
            .body(request);

    // Send a request
    MobageBankDebitPostResponse response;
    try {
      ResponseEntity<MobageBankDebitPostResponse> responseEntity =
          this.restTemplate.exchange(requestEntity, MobageBankDebitPostResponse.class);
      response = responseEntity.getBody();
    } catch (HttpClientErrorException e) {

      // In this case, the error is caused by Invalid Request
      // Status Code 400； The error is caused by invalid parameter in JSON Request Body
      // Status Code 401； The error is caused by invalid HTTP Authorization Header
      //                   Please make sure if Authorization Header is valid.
      logger.error(e.getResponseBodyAsString());
      if (e.getStatusCode() == HttpStatus.UNAUTHORIZED) {
        throw new MobageUnauthorizedException(
            "Failed to communitate to Mobage because valid access token is not set");
      } else {
        throw new MobageException(
            "Failed to communicate to Mobage because of an invalid request. Status Code:"
                + e.getStatusCode());
      }
    } catch (HttpServerErrorException e) {
      // In this case, the error might be caused by Mobage Platform
      throw new MobageException("Unexpected error occurred on Mobage Platform");
    } catch (RestClientException e) {
      // Typically in case of error while extracting response
      throw new MobageException(e.getMessage());
    }
    return response;
  }

  /**
   * Send a request to Mobage Server and get the payment process.
   */
  private MobageBankDebitGetResponse getBankDebit(String transactionId, String accessToken) {

    // Set authorization header with Access Token
    HttpHeaders headers = new HttpHeaders();
    headers.add("Authorization", "Bearer " + accessToken);
    headers.add("Accept", "*/*");

    // Send a request
    MobageBankDebitGetResponse response;
    RestTemplate restTemplate = new RestTemplate();
    try {
      String url = config.getBankDebitEndpoint() + "/" + transactionId;
      ResponseEntity<MobageBankDebitGetResponse> responseEntity =
          restTemplate.exchange(
              RequestEntity.get(URI.create(url)).headers(headers).build(),
              MobageBankDebitGetResponse.class);
      response = responseEntity.getBody();
    } catch (HttpClientErrorException e) {

      // In this case, the error is caused by Invalid Request
      // Status Code 400； The error is caused by invalid parameter in JSON Request Body
      // Status Code 401； The error is caused by invalid HTTP Authorization Header
      //                   Please make sure if Authorization Header is valid.
      logger.error(e.getResponseBodyAsString(), e);
      if (e.getStatusCode() == HttpStatus.UNAUTHORIZED) {
        throw new MobageUnauthorizedException(
            "Failed to communitate to Mobage "
                + "because valid access token is not set in HttpSession");
      } else {
        throw new MobageException(
            "Failed to communicate to Mobage because of an invalid request. Status Code:"
                + e.getStatusCode());
      }
    } catch (HttpServerErrorException e) {
      // In this case, the error might be caused by Mobage Platform
      throw new MobageException("Unexpected error occurred on Mobage Platform");
    } catch (RestClientException e) {
      // Typically in case of error while extracting response
      throw new MobageException(e.getMessage());
    }
    return response;
  }

  /**
   * Publish orderId and start order process.
   */
  public String publishOrderId(String userId, String transactionId) {
    int orderId = 0;

    final PreparedStatementCreator psc =
        new PreparedStatementCreator() {
          @Override
          public PreparedStatement createPreparedStatement(final Connection connection)
              throws SQLException {
            final PreparedStatement stmt =
                connection.prepareStatement(
                    "insert ignore into order_db"
                        + " (transaction_id, user_id, client_id, created_at) values(?, ?, ?, ?)",
                    Statement.RETURN_GENERATED_KEYS);
            stmt.setString(1, transactionId);
            stmt.setString(2, userId);
            stmt.setString(3, config.getClientId());
            stmt.setLong(4, System.currentTimeMillis() / 1000);

            return stmt;
          }
        };
    final KeyHolder holder = new GeneratedKeyHolder();
    jdbcTemplate.update(psc, holder);
    // Get the last inserted orderId
    orderId = holder.getKey().intValue();
    return Integer.toString(orderId);
  }

  /**
   * Select userId from orderId in database
   */
  public String selectUserIdFromOrderId(String orderId) {
    Map<String, Object> user =
        jdbcTemplate.queryForMap("select user_id from order_db where order_id = ?", orderId);
    return user.get("user_id").toString();
  }

  /**
   * Give items and close order process.
   **/
  @Transactional
  public void giveOrderItem(String orderId, String itemId, int quantity) {

    Map<String, Object> order =
        jdbcTemplate.queryForMap("select * from order_db where order_id = ? for update", orderId);

    if (!order.get("order_state").toString().equals("authorized")) {
      throw new MobageException("order_state is not authorized.");
    }
    String userId = order.get("user_id").toString();
    jdbcTemplate.update(
        "insert into user_items"
            + " (user_id, item_id, item_num) values (?, ?, ?)"
            + " on duplicate key update item_num = item_num + ?;",
        userId,
        itemId,
        quantity,
        quantity);

    jdbcTemplate.update("update order_db set order_state = ? where order_id = ?", "close", orderId);
  }

  /**
   *  Cancel order process.
   **/
  @Transactional
  public void cancelOrder(String orderId, String transactionId, String accessToken) {

    Map<String, Object> order =
        jdbcTemplate.queryForMap("select * from order_db where order_id = ? for update", orderId);
    if (!order.get("order_state").toString().equals("authorized")) {
      throw new MobageException("order_state is not authorized.");
    }
    if (!order.get("transaction_id").toString().equals(transactionId)) {
      throw new MobageException("transaction_id is invalid.");
    }
    String userId = order.get("user_id").toString();

    MobageBankDebitGetResponse res = getBankDebit(transactionId, accessToken);
    String state = res.getState();
    if (state == null) {
      logger.error(
          String.format(
              "Failure. Not Found. {order_id: %s, user_id: %s, %s}",
              orderId, userId, res.getItems().get(0).getItem().getId()));
    } else if (state.equals("canceled") || state.equals("error")) {
      cancelOrderItem(orderId, state);
      logger.info(String.format("Canceled. {order_id: %s, user_id: %s}", orderId, userId));
    } else {
      logger.error(
          String.format(
              "Failure. Platform Transaction state is not \"canceled\" or \"error\". {order_id: %s, user_id: %s}",
              orderId, userId));
    }
  }

  private void cancelOrderItem(String orderId, String state) {
    jdbcTemplate.update("update order_db set order_state = ? where order_id = ?;", state, orderId);
  }

}
